#include "PLA.h"
#include "mpi.h"

static int FIRST_WARNING = 1;
static double PLA_DUMMY_ADRESS;

#if MANUFACTURE == CRAY
  #define PLA_CRAY_NULL_FIX( name ) ( ( name ) ? ( name ) : 1 )
  #define BF(buffer) (NULL==buffer) ? &PLA_DUMMY_ADRESS : buffer

  #define PLA_MPI_Allgatherv(a,b,c,d,e,f,g,h)   \
       do {                                     \
          MPI_Barrier( h );                     \
          MPI_Allgatherv(a,b,c,d,e,f,g,h);      \
       } while(0)

#else
  #define PLA_CRAY_NULL_FIX( name )  name
  #define BF(buffer) buffer
  #define PLA_MPI_Allgatherv(a,b,c,d,e,f,g,h)   \
          MPI_Allgatherv    (a,b,c,d,e,f,g,h)
#endif



int Copy_mscalar_to_mscalar( PLA_Obj from, PLA_Obj to )
{
  PLA_Obj duplicate = NULL;
  PLA_Template template = NULL;
  MPI_Comm comm_col, comm_row, comm_all;
  int myrow, mycol, myall, 
      nprows, npcols, npall;
  int mm,nn;
  int owner_col_from, owner_row_from;
  int m, n, to_ldim;
  int bcast_row = 0, bcast_col = 0, bcast_root;
  char *buf;
  void *duplicate_buf = NULL, *to_buf = NULL;
  MPI_Datatype datatype;

  /*  Copy_mscalar_to_mscalar_enter( from, to ); */

  PLA_Obj_template( from, &template );
  PLA_Temp_comm_col_info( template, &comm_col, &myrow, &nprows );
  PLA_Temp_comm_row_info( template, &comm_row, &mycol, &npcols );
  PLA_Temp_comm_all_info( template, &comm_all, &myall, &npall);

  PLA_Obj_owner_col( from, &owner_col_from );
  PLA_Obj_owner_row( from, &owner_row_from );

  PLA_Obj_global_length ( from, &m );
  PLA_Obj_global_width  ( from, &n );
  PLA_Obj_datatype      ( from, &datatype );  




  /* if the _from mscalar is not duplicated to all nodes */
  if ( (owner_col_from != PLA_ALL_COLS) || 
       (owner_row_from != PLA_ALL_ROWS) )
    {
      /* first we need to duplicate the multiscalar _from */
      PLA_Mscalar_create( datatype, 
			  PLA_ALL_ROWS, PLA_ALL_COLS,
			  m, n,
			  template,
			  &duplicate );
      PLA_Obj_local_buffer( duplicate, &duplicate_buf );
      
      /* where shall we bcast from ? */
      if (owner_col_from != PLA_ALL_COLS) bcast_col = owner_col_from;
      if (owner_row_from != PLA_ALL_ROWS) bcast_row = owner_row_from;
      bcast_root = bcast_col * nprows + bcast_row;

      /* bcast to all nodes, this could be redundant at times */
      if( myall == bcast_root ){

	PLA_Obj_get_local_contents( from, 
				    PLA_NO_TRANSPOSE, 
				    &mm, &nn, 
				    duplicate_buf, m, 1 );
      }
      MPI_Bcast( duplicate_buf, m*n, datatype, bcast_root, comm_all );
      
    }
  else /* the from multiscalar is allready duplicated */
    {
      duplicate = from;
    }
  /* now copy duplicate into _to_ */
  PLA_Obj_local_buffer( to, &to_buf );
  PLA_Obj_local_ldim ( to, &to_ldim );
  if ( to_buf != NULL )
    PLA_Obj_get_local_contents( duplicate, 
				PLA_NO_TRANSPOSE, 
				&mm, &nn, 
				to_buf, to_ldim, 1 );
      

  if ( (owner_col_from != PLA_ALL_COLS) || 
       (owner_row_from != PLA_ALL_ROWS) )
       PLA_Obj_free( &duplicate );
  return(0);
}



int PLA_Copy_mvector_to_mscalar( PLA_Obj mv, PLA_Obj ms)
{
  /* input : multivector _mv_ 
             multiscalar _ms_ 
	     global sizes of the input objects should match for
	     the successful completion of the operation
     output: 

     here is the plan:

     1) create the temporary multiscalar duplicated on all nodes
        and conformal to the input mscalar.
     2) perform an all_gather into the temporary buffer.
     3) Unravel the temporary buffer into temporary multiscalar duplicated
        ON ALL NODES
     4) copy temporary mscalar into output multiscalar

  */

  PLA_Obj recv = NULL, temp_ms = NULL, recv_view = NULL, temp_view = NULL;
  PLA_Obj temp=NULL;
  PLA_Template template = NULL;
  MPI_Comm comm_all;
  int myrow, mycol, myall, 
      nprows, npcols, npall;
  int mv_length, mv_width, mv_align;
  int align, oldalign, nb;
  int mm,nn, lld;
  int top_owner, top_size, shift_size;
  int i, total=0, count = 0;
  int firsttime = 1;
  int *recvcounts = NULL, *displacement = NULL, *displ = NULL;
  int bcast_row = 0, bcast_col = 0, bcast_root;
  void *recv_buffer = NULL, *mv_buffer = NULL, *temp_ms_buffer = NULL;
  void *send_buffer = NULL, *temp_buffer = NULL, *temp_view_buffer = NULL;
  MPI_Datatype datatype;

  /*  Copy_mvector_to_mscalar_enter( mv, ms ); */
  PLA_Obj_template( mv, &template );
  PLA_Temp_comm_all_info( template, &comm_all, &myall, &npall);
  PLA_Obj_datatype( mv, &datatype);
  PLA_Obj_global_length(mv, &mv_length);
  PLA_Obj_global_width(mv, &mv_width);
  PLA_Obj_global_align_row(mv, &mv_align);
  PLA_Obj_local_buffer(mv, &mv_buffer);
  PLA_Temp_nb( template, &nb );

  displacement = (int *) malloc( npall * sizeof(int) );
  displacement[0] = 0;
  displ = (int *) malloc( npall * sizeof(int) );
  displ[0] = 0;
  recvcounts = (int *) malloc( npall * sizeof(int) );
  PLA_Temp_vector_distant_length( template, 0,
				  mv_length, mv_align, 
				  &recvcounts[0]);
  recvcounts[0] *= mv_width;
  for(i=1; i<npall; i++)
    {
      /* find number of rows communicated by the i-th node */
      PLA_Temp_vector_distant_length( template, i, 
				      mv_length, mv_align, 
				      &recvcounts[i]);
      recvcounts[i] *= mv_width;
      /* this column's data begins just after the end of the last column's */ 
      displacement[ i ] = displacement[ i-1] + recvcounts[i-1];
      displ[i] = displacement[i];
    }

  PLA_Mscalar_create_conf_to( ms, PLA_ALL_ROWS, PLA_ALL_COLS, &recv );
  PLA_Obj_local_buffer( recv, &recv_buffer);

  PLA_Mscalar_create( datatype, 
		      PLA_ALL_ROWS, PLA_ALL_COLS, 
		      mv_width, mv_length, template,
		      &temp_ms );
  PLA_Obj_local_buffer( temp_ms, &temp_ms_buffer );

  PLA_Mscalar_create_conf_to( ms, PLA_ALL_ROWS, PLA_ALL_COLS, &temp );
  PLA_Obj_local_buffer( temp, &temp_buffer );

  PLA_Obj_get_local_contents( mv, 
			      PLA_TRANSPOSE, 
			      &mm, &nn, 
			      temp_buffer, mv_width, 1 );
  
  MPI_Allgatherv( temp_buffer, recvcounts[ myall ], datatype,
		  temp_ms_buffer, recvcounts, displacement,
		  datatype, comm_all);
  
  PLA_Obj_get_local_contents( temp_ms, 
			      PLA_TRANSPOSE, 
			      &mm, &nn, 
			      recv_buffer, mv_length, 1 );

  /* unshuffle recv_buffer into temp_ms */

  align = displacement[0];
  PLA_Obj_top( mv, &top_size, &top_owner);
  shift_size = top_size;

  for(;;){
    if(total >= mv_length) break;
    oldalign = align;
    align = displacement[top_owner]/mv_width;

        if(firsttime)
	  {
	    PLA_Mscalar_submscalar( recv, 
				    top_size, mv_width,
				    align, PLA_ALIGN_FIRST,
				    &recv_view );


	    PLA_Mscalar_submscalar( temp,
				    top_size, mv_width,
				    PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
				    &temp_view );
	    firsttime = 0; 
	  }
	else
	  {
	    /* what if the final chunk is shorter than nb? */
            /* jwu quick fix */
	    if (recvcounts[top_owner]/mv_width < nb)
	      nb=recvcounts[top_owner]/mv_width;

	    PLA_Mscalar_view_shift( temp_view,        shift_size,
				                0,                    0,
                                                            nb              );
	    PLA_Mscalar_view_shift( recv_view,         align-oldalign,      
				                0,                        0,
				               align-oldalign+nb-shift_size );

	    shift_size = nb;

	  }

	PLA_Obj_local_buffer(temp_view, &temp_view_buffer);
	PLA_Obj_local_ldim( temp_view, &lld );

	PLA_Obj_get_local_contents( recv_view, 
				    PLA_NO_TRANSPOSE, 
				    &mm, &nn, 
				    temp_view_buffer, lld, 1 );
	displacement[top_owner] += shift_size*mv_width;
        /* jwu quick fix */
	recvcounts[top_owner] -= shift_size*mv_width;
	total+=shift_size;
	top_owner = (top_owner+1)%npall;
  }
  PLA_Copy( temp, ms);

  free(recvcounts); free(displ); free(displacement);
  PLA_Obj_free( &recv_view ); PLA_Obj_free( &temp_view );
  PLA_Obj_free( &recv ); PLA_Obj_free( &temp_ms );
  PLA_Obj_free ( & temp );
  return(0);
}




int PLA_Copy_mscalar_to_mvector( PLA_Obj ms, PLA_Obj mv )
{
  /* input : multivector _mv_ 
             multiscalar _ms_ 
	     global sizes of the input objects should match for
	     the successful completion of the operation
     output: 

     here is the plan:

     1) create the temporary multivector duplicated on all nodes
        and conformal to the input mscalar.
     2) perform an all_gather into the temporary buffer.
     3) Unravel the temporary buffer into temporary multiscalar duplicated
        ON ALL NODES
     4) copy temporary mscalar into 

  */

  PLA_Obj send = NULL, ms_view = NULL, send_view = NULL, recv = NULL;
  PLA_Obj recv_view = NULL;
  PLA_Template template = NULL;
  MPI_Comm comm_all;
  int myrow, mycol, myall, 
      nprows, npcols, npall;
  int mv_length, mv_width, mv_align;
  int mv_loc_max;
  int align, oldalign, nb;
  int mm,nn, lld, mv_lld;
  int top_owner, top_size, shift_size;
  int owner_col, owner_row;
  int i, total=0;
  int firsttime = 1;
  int root, root_col=0, root_row=0;
  int *sendcounts = NULL, *displacements = NULL, *displ = NULL, *jwu=NULL;
  int bcast_row = 0, bcast_col = 0, bcast_root;
  void *send_view_buffer = NULL, *mv_buffer = NULL, *send_buffer=NULL;
  void *recv_buffer=NULL;
  MPI_Datatype datatype;

  /*  Copy_mvector_to_mscalar_enter( mv, ms ); */

  PLA_Obj_template( mv, &template );
  PLA_Temp_comm_all_info( template, &comm_all, &myall, &npall);
  PLA_Obj_datatype( mv, &datatype);
  PLA_Obj_global_length(mv, &mv_length);
  PLA_Obj_global_width(mv, &mv_width);
  PLA_Obj_global_align_row(mv, &mv_align);
  PLA_Obj_local_buffer(mv, &mv_buffer);
  PLA_Temp_nb( template, &nb );
  PLA_Temp_comm_col_size( template, &nprows);

  PLA_Obj_owner_col( ms, &owner_col );
  PLA_Obj_owner_row( ms, &owner_row );


  /* where shall we scatter from ? */
  if (owner_col != PLA_ALL_COLS) 
    root_col = owner_col;
  else
    root_col = 0;
  if (owner_row != PLA_ALL_ROWS) 
    root_row = owner_row;
  else
    root_row = 0;

  root = root_col * nprows + root_row;

  /* create the multiscalar buffer from which the scatter would be
     perforemed */
  PLA_Mscalar_create( datatype, owner_row, owner_col,
		      mv_width, mv_length, template, &send );
  PLA_Obj_local_buffer( send, &send_buffer);

  /* compute the parameters of PLA_Scatter_v */
  displacements = (int *) malloc( npall * sizeof(int) );
  displacements[0] = 0;
  displ = (int *) malloc( npall * sizeof(int) );
  displ[0] = 0;

  sendcounts = (int *) malloc( npall * sizeof(int) );
  jwu = (int *) malloc( npall * sizeof(int) );
  PLA_Temp_vector_distant_length( template, 0,
				  mv_length, mv_align, 
				  &sendcounts[0]);
  jwu[0] = sendcounts[0]; sendcounts[0] *= mv_width;  

  for(i=1; i<npall; i++)
    {
      /* find number of rows communicated by the i-th node */
      PLA_Temp_vector_distant_length( template, i, 
				      mv_length, mv_align, 
				      &sendcounts[i]);
      jwu[i] = sendcounts[i]; sendcounts[i] *= mv_width; 

      /* this column's data begins just after the end of the last column's */ 
      displacements[ i ] = displacements[ i-1] + sendcounts[i-1];
      displ[i] = displacements[i];
    }

  mv_loc_max = ((mv_length/nb + 1)/npall + 1) * nb; 
  PLA_Mscalar_create(datatype,
		     PLA_ALL_ROWS, PLA_ALL_COLS,
		     mv_width,mv_loc_max,
		     template,
		     &recv); 

  /*  PLA_Mvector_create_conf_to( mv, mv_width, &recv ); */
  PLA_Obj_local_buffer( recv, &recv_buffer);

  /* shuffle ms_buffer into send_buffer */
  PLA_Obj_top( mv, &top_size, &top_owner);
  shift_size = top_size;

  for(;;){
    if(total >= mv_length) break;
    oldalign = align;
    align = displacements[top_owner]/mv_width;
        if(firsttime)
	  {
	    PLA_Mscalar_submscalar( ms, 
				    top_size, mv_width,
				    PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
				    &ms_view );

	    PLA_Mscalar_submscalar( send,
				    top_size, mv_width,
				    align*mv_width, PLA_ALIGN_FIRST,
				    &send_view );
	    firsttime = 0;
	  }
	else
	  {
	    /* what if the final chunk is shorter than nb? */
            if ( jwu[top_owner] < nb )
              nb = jwu[top_owner];

 	    PLA_Mscalar_view_shift( ms_view,        shift_size,
				                0,                    0,
                                                            nb              );
	    PLA_Mscalar_view_shift( send_view,    0,   align-oldalign,      
 				            align-oldalign+nb-shift_size,  0);
	    shift_size = nb;
	  }

	PLA_Obj_local_buffer(send_view, &send_view_buffer);
	PLA_Obj_local_ldim( send_view, &lld );

	PLA_Obj_get_local_contents( ms_view, 
				    PLA_TRANSPOSE, 
				    &mm, &nn, 
				    send_view_buffer, lld, 1 );
	displacements[top_owner] += shift_size*mv_width;
	jwu[top_owner] -= shift_size;
	top_owner = (top_owner+1)%npall;
	total+=shift_size;
  }

  MPI_Scatterv( send_buffer, sendcounts,  displ, datatype,
		recv_buffer, sendcounts[myall], 
		datatype, root, comm_all);

  PLA_Obj_local_ldim( mv, &mv_lld );

  for(i=0;i<npall;i++){
    if(i==myall)
      PLA_Mscalar_submscalar( recv, 
			      mv_width, sendcounts[myall]/mv_width,
			      PLA_ALIGN_FIRST, PLA_ALIGN_FIRST,
			      &recv_view);
  }

  PLA_Obj_get_local_contents( recv_view,
			      PLA_TRANSPOSE,
			      &mm, &nn,
			      mv_buffer, mv_lld,1);


  free(jwu); free(sendcounts); free(displ); free(displacements);

  PLA_Obj_free( &send_view ); PLA_Obj_free( &ms_view );
  PLA_Obj_free( &send ); PLA_Obj_free( &recv_view ); 
  PLA_Obj_free( &recv ); 
  return(0);
}


int Copy_matrix_to_mscalar( PLA_Obj mat, PLA_Obj ms )
{
  /* matrix into multiscalar copy */
  int
    orientation,
    mv_width;
  PLA_Obj
    temp = NULL;

  PLA_Obj_project_onto( mat, &orientation);
  if (orientation == PLA_PROJ_ONTO_COL)
    PLA_Obj_global_width( mat, &mv_width);
  else
    PLA_Obj_global_length( mat, &mv_width);

  PLA_Mvector_create_conf_to( mat, mv_width, &temp);
  PLA_Copy( mat, temp );

  PLA_Copy_mvector_to_mscalar( temp, ms );

  PLA_Obj_free ( & temp );

  return(0);
}

			     

int Copy_mscalar_to_matrix( PLA_Obj ms, PLA_Obj mat)
{
  int
    orientation,
    mv_width;
  PLA_Obj
    temp = NULL;

  PLA_Obj_project_onto( mat, &orientation);
  if (orientation == PLA_PROJ_ONTO_COL)
    PLA_Obj_global_width( mat, &mv_width);
  else
    PLA_Obj_global_length( mat, &mv_width);
    
  PLA_Mvector_create_conf_to( mat, mv_width, &temp);

  PLA_Copy_mscalar_to_mvector( ms, temp );
  PLA_Copy( temp, mat );

  PLA_Obj_free( &temp );
  return(0);
}
