#include "mpi.h"
#include "mmpi.h"
#include "SL.h"
#include "MISC_DEFS.h"
#include "PLA.MMPI.h"
#include "PLA_LAObj_open_close.h"

int PLA0_LAObj_open( int         mode,
                     SL_LAObj    laobj )
/* 
 *  PLA_LAObj_open 
 *
 *  Purpose: open LAObj object for reading or writing
 */
{
  SL_LAObj_obj_type       objtype;
  MPI_Datatype            datatype;
  MPI_Comm                comm;
  SL_Template             template;
  struct PLA0_Fill_info  *laobj_fill;
  extern int              PLA0_LAObj_write_consumer(),
                          PLA0_LAObj_read_consumer(),
                          PLA0_LAObj_read_bouncer();

  /* check object type */
  SL_LAObj_objtype( laobj, &objtype );
  if ( objtype != SL_VECTOR && objtype != SL_MATRIX ) {
    printf("Attempt to read or write object other than vector or matrix.\n");
    exit( 0 );
  };

  /* check if object is already open */
  if ( laobj->fill != NULL ) {
    printf("Attempt to open object that is already open.\n");
    exit( 0 );
  }

  /* get the fill field */
  laobj_fill = (struct PLA0_Fill_info *)
                   malloc( sizeof( struct PLA0_Fill_info ) );
  laobj->fill = laobj_fill;
   
  /* Set the appropriate fields */
  if ( mode != PLA0_READ && mode != PLA0_WRITE ) {
    printf("Attempt to open object with unknown mode\n");
    exit( 0 );
  }

  laobj_fill->mode             = mode;
  laobj_fill->consumer_receive = MMPI_CON_NULL;
  laobj_fill->consumer_bounce  = MMPI_CON_NULL;
  laobj_fill->buffer           = MMPI_BUF_NULL;

  /* get communicator */
  SL_LAObj_template( laobj, &template );
  SL_Temp_comm1D( template, &comm );

  /* barrier, to make sure all nodes are opening object */
  MMPI_Barrier( comm );

  /* create consumers */
  if ( mode == PLA0_WRITE ) {
    /* create a write consumer */
    MMPI_Con_create( comm, laobj, PLA0_LAObj_write_consumer,
		     &laobj_fill->consumer_receive );
  }
  else {
    /* create the read bounce consumer */
    MMPI_Con_create( comm, laobj, PLA0_LAObj_read_consumer,
		     &laobj_fill->consumer_receive );
    /* create the read bounce consumer */
    MMPI_Con_create( comm, laobj, PLA0_LAObj_read_bouncer,
		     &laobj_fill->consumer_bounce );
  }
}


int PLA0_LAObj_close( SL_LAObj laobj )
/* 
 *  PLA_LAObj_close
 *
 *  Purpose: close LAObj for reading and writing
 */
{
  struct PLA0_Fill_info *laobj_fill;
  SL_Template template = NULL;
  MPI_Comm comm = NULL;

  /* check if object is open for filling */
  if ( laobj->fill == NULL ) {
    printf("Attempting to close object that is not open \n");
    exit( 0 );
  }

  laobj_fill = ( struct PLA0_Fill_info * ) laobj->fill;

  SL_LAObj_template( laobj, &template );
  SL_Temp_comm1D( template, &comm );
  MMPI_Barrier( comm );

  /* free the consumer and buffer space */
  if ( laobj_fill->consumer_bounce != NULL ) 
    MMPI_Con_free( &laobj_fill->consumer_bounce );
  MMPI_Barrier( comm );
  MMPI_Con_free( &laobj_fill->consumer_receive );
  MMPI_Barrier( comm );
  if(laobj_fill->buffer != NULL) /* JAG patch -- necessary but why? */
     MMPI_Buf_free( &laobj_fill->buffer );

  free( laobj_fill );
  laobj->fill = NULL;
}


