/* ---------------------------------------------------------------------
Author:     H. Carter Edwards
		http://www.ticam.utexas.edu/~carter
		carter@ticam.utexas.edu

Copyright:  Copyright (C) 1996   H. Carter Edwards

Purpose:    Asynchronous message management for MPI

License:    See 'License.Gnu'

Mail:
	University of Texas at Austin
	TICAM, TAY 2.400
	2313 Speedway
	Austin, TX  78712

Disclaimer:

    MMPI comes with ABSOLUTELY NO WARRANTY; for details read
    the file License.Gnu that came with this library.
    This is free software, and you are welcome to redistribute it
    under certain conditions, i.e. read the file 'License.Gnu'.
--------------------------------------------------------------------- */

#include "mmpi.h"
#include <memory.h>
#include <stdlib.h>

struct MMPI_Buf_type {
  MPI_Comm comm ;
  int      maxsize ;
  int      size ;
  int      pos ;
  void   * base ;
};

/* ------------------------------------------------------------------- */

int MMPI_Buf_free( MMPI_Buf * buffer )
{
  const char my_name[] = "MMPI_Buf_free" ;

  if ( buffer == (MMPI_Buf *) NULL || *buffer == MMPI_BUF_NULL ) {
    return MPI_ERR_ARG ;
  }

  if ( (*buffer)->base !=
       (void *)(((char*)*buffer)+sizeof(struct MMPI_Buf_type))) {
    fprintf(MMPI_Log_file(),"%s : corrupted buffer",my_name);
    fflush(MMPI_Log_file());
    return MPI_ERR_BUFFER ;
  }

  free( *buffer );

  *buffer = NULL ;

  return MPI_SUCCESS ;
}

/* ------------------------------------------------------------------- */

int MMPI_Buf_create( int size , MPI_Comm comm , MMPI_Buf * buffer )
{
  const char my_name[] = "MMPI_Buf_create" ;

  if ( buffer == (MMPI_Buf *) NULL ) {
    fprintf(MMPI_Log_file(),"%s : Null pointer", my_name);
    fflush(MMPI_Log_file());
    return MPI_ERR_ARG ;
  }

  size = ( size < 0 ) ? 0 : size ;

  *buffer = (MMPI_Buf) malloc( size + sizeof(struct MMPI_Buf_type) );

  if ( NULL == *buffer ) {
    fprintf(MMPI_Log_file(),"%s : Failed malloc", my_name);
    fflush(MMPI_Log_file());
    return MPI_ERR_BUFFER ;
  }

  (*buffer)->comm = comm ;
  (*buffer)->pos  = 0 ;
  (*buffer)->size = 0 ;
  (*buffer)->base = ((char *)(*buffer)) + sizeof(struct MMPI_Buf_type) ;
  (*buffer)->maxsize = size ;

  return MPI_SUCCESS ;
}

/* ------------------------------------------------------------------- */

int MMPI_Buf_copy( MMPI_Buf old_buf , MMPI_Buf * new_buf )
{
  int R ;

  if ( MPI_SUCCESS != ( R =
    MMPI_Buf_create( old_buf->size , old_buf->comm , new_buf ) ) ) {
    return R ;
  }

  (*new_buf)->pos  = old_buf->pos ;
  (*new_buf)->size = old_buf->size ;

  if ( old_buf->size )
    memcpy( (*new_buf)->base , old_buf->base , old_buf->size );

  return MPI_SUCCESS ;
}

/* ------------------------------------------------------------------- */

int MMPI_Buf_capacity( MMPI_Buf buffer , int * size )
{
  if ((buffer)->base != (void *)(((char*)buffer)+sizeof(struct MMPI_Buf_type))){
    return MPI_ERR_BUFFER ;
  }
  *size = (buffer)->maxsize ;

  return MPI_SUCCESS ;
}

/* ------------------------------------------------------------------- */

int MMPI_Buf_remain( MMPI_Buf buffer , int * rem )
{
  if ((buffer)->base!=(void *)(((char*)buffer)+sizeof(struct MMPI_Buf_type))){
    return MPI_ERR_BUFFER ;
  }
  *rem = (buffer)->size - (buffer)->pos ;

  return MPI_SUCCESS ;
}

/* ------------------------------------------------------------------- */

int MMPI_Buf_position( MMPI_Buf buffer , int * pos )
{
  if ((buffer)->base != (void *)(((char*)buffer)+sizeof(struct MMPI_Buf_type))){
    return MPI_ERR_BUFFER ;
  }
  *pos = (buffer)->pos ;

  return MPI_SUCCESS ;
}

/* ------------------------------------------------------------------- */

int MMPI_Buf_pointer( MMPI_Buf buffer , void ** ptr )
{
  if ((buffer)->base != (void *)(((char*)buffer)+sizeof(struct MMPI_Buf_type))){
    return MPI_ERR_BUFFER ;
  }
  *ptr = (buffer)->base ;

  return MPI_SUCCESS ;
}

/* ------------------------------------------------------------------- */

int MMPI_Buf_reset( int size , MPI_Comm Comm , MMPI_Buf* buffer )
{
  if ((*buffer)->base!=(void *)(((char*)*buffer)+sizeof(struct MMPI_Buf_type))){
    return MPI_ERR_BUFFER ;
  }

  if ( (*buffer)->maxsize < size ) {
    int newsize = sizeof(struct MMPI_Buf_type) + size ;

    *buffer = (MMPI_Buf) realloc( (void *) *buffer , newsize );

    if ( NULL == *buffer ) return MPI_ERR_BUFFER ;

    (*buffer)->base = ((char *)*buffer) + sizeof(struct MMPI_Buf_type) ;
    (*buffer)->maxsize = size ;
  }

  (*buffer)->comm = Comm ;
  (*buffer)->size = size ;
  (*buffer)->pos  = 0 ;

  return MPI_SUCCESS ;
}

/* ------------------------------------------------------------------- */

int MMPI_Buf_size( MMPI_Buf buffer , int * size )
{
  if ((buffer)->base!=(void *)(((char*)buffer)+sizeof(struct MMPI_Buf_type))){
    return MPI_ERR_BUFFER ;
  }
  *size = (buffer)->size ;

  return MPI_SUCCESS ;
}

/* ------------------------------------------------------------------- */

int MMPI_Buf_comm( MMPI_Buf buffer , MPI_Comm * Comm )
{
  if ((buffer)->base!=(void *)(((char*)buffer)+sizeof(struct MMPI_Buf_type))){
    return MPI_ERR_BUFFER ;
  }
  *Comm = (buffer)->comm ;

  return MPI_SUCCESS ;
}

/* ------------------------------------------------------------------- */

MMPI_Buf_unpack(
  MMPI_Buf buffer ,
  void          * outbuf ,
  int             outcount ,
  MPI_Datatype    datatype )
{
  if ((buffer)->base!=(void *)(((char*)buffer)+sizeof(struct MMPI_Buf_type))){
    return MPI_ERR_BUFFER ;
  }

  return MPI_Unpack(
    buffer->base, buffer->size, &(buffer->pos),
    outbuf , outcount , datatype, buffer->comm );
}

/* ------------------------------------------------------------------- */

MMPI_Buf_pack(
  void           * inbuf ,
  int              incount ,
  MPI_Datatype     datatype ,
  MMPI_Buf* buffer )
{
  int R ;
  int size ;

  if ((*buffer)->base!=(void *)(((char*)*buffer)+sizeof(struct MMPI_Buf_type))){
    return MPI_ERR_BUFFER ;
  }

  /* Size of this message */

  if ( MPI_SUCCESS != (
    R = MPI_Pack_size( incount, datatype, (*buffer)->comm, &size) ) ) {
    return R ;
  }

  if ( (*buffer)->maxsize < (*buffer)->pos + size ) {

    int newsize = sizeof(struct MMPI_Buf_type) + (*buffer)->pos + size ;

    *buffer = (MMPI_Buf) realloc( (void *) *buffer , newsize );

    if ( NULL == *buffer ) return MPI_ERR_BUFFER ;

    (*buffer)->base = ((char *)*buffer) + sizeof(struct MMPI_Buf_type) ;

    (*buffer)->maxsize = newsize - sizeof(struct MMPI_Buf_type);
  }

  return MPI_Pack( inbuf, incount, datatype,
    (*buffer)->base, (*buffer)->maxsize, &((*buffer)->pos), (*buffer)->comm );
}

/* ------------------------------------------------------------------- */

int MMPI_Buf_status( MMPI_Buf buf , MPI_Status * status )
{
  int R ;
  int count ;

  if ( (buf)->base!=(void *)(((char*)buf)+sizeof(struct MMPI_Buf_type))){
    return MPI_ERR_BUFFER ;
  }

  if ( MPI_SUCCESS != ( R = MPI_Get_count( status, MPI_PACKED, &count ) ) ) {
    return R ;
  }

  if ( buf->maxsize < count ) {
    return MPI_ERR_COUNT ;
  }

  buf->size = count ;
  buf->pos = 0 ;

  return MPI_SUCCESS ;
}

int MMPI_Buf_recv(  MMPI_Buf buf, int src , int tag , MPI_Status * status )
{
  int R ;

  if ( (buf)->base!=(void *)(((char*)buf)+sizeof(struct MMPI_Buf_type))){
    return MPI_ERR_BUFFER ;
  }

  return MPI_Recv( buf->base, buf->maxsize, MPI_PACKED,
    src, tag, buf->comm, status );
}

int MMPI_Buf_send(  MMPI_Buf buf, int dest, int tag )
{
  int R ;

  if ( (buf)->base!=(void *)(((char*)buf)+sizeof(struct MMPI_Buf_type))){
    return MPI_ERR_BUFFER ;
  }

  return MPI_Send( buf->base, buf->pos, MPI_PACKED, dest, tag, buf->comm );
}

int MMPI_Buf_rsend( MMPI_Buf buf, int dest, int tag )
{
  int R ;

  if ( (buf)->base!=(void *)(((char*)buf)+sizeof(struct MMPI_Buf_type))){
    return MPI_ERR_BUFFER ;
  }

  return MPI_Rsend( buf->base, buf->pos, MPI_PACKED, dest, tag, buf->comm );
}

int MMPI_Buf_ssend( MMPI_Buf buf, int dest, int tag )
{
  int R ;

  if ( (buf)->base!=(void *)(((char*)buf)+sizeof(struct MMPI_Buf_type))){
    return MPI_ERR_BUFFER ;
  }

  return MPI_Ssend( buf->base, buf->pos, MPI_PACKED, dest, tag, buf->comm );
}

int MMPI_Buf_bsend( MMPI_Buf buf, int dest, int tag )
{
  int R ;

  if ( (buf)->base!=(void *)(((char*)buf)+sizeof(struct MMPI_Buf_type))){
    return MPI_ERR_BUFFER ;
  }

  return MPI_Bsend( buf->base, buf->pos, MPI_PACKED, dest, tag, buf->comm );
}


int MMPI_Buf_irecv(  MMPI_Buf buf, int src,  int tag, MPI_Request * Req )
{
  int R ;

  if ( (buf)->base!=(void *)(((char*)buf)+sizeof(struct MMPI_Buf_type))){
    return MPI_ERR_BUFFER ;
  }

  return MPI_Irecv( buf->base, buf->maxsize, MPI_PACKED,
    src, tag, buf->comm, Req );
}

int MMPI_Buf_isend(  MMPI_Buf buf, int dest, int tag, MPI_Request * Req )
{
  int R ;

  if ( (buf)->base!=(void *)(((char*)buf)+sizeof(struct MMPI_Buf_type))){
    return MPI_ERR_BUFFER ;
  }

  return MPI_Isend( buf->base, buf->pos, MPI_PACKED, dest, tag, buf->comm, Req);
}

int MMPI_Buf_irsend( MMPI_Buf buf, int dest, int tag, MPI_Request * Req )
{
  int R ;

  if ( (buf)->base!=(void *)(((char*)buf)+sizeof(struct MMPI_Buf_type))){
    return MPI_ERR_BUFFER ;
  }

  return MPI_Irsend( buf->base, buf->pos, MPI_PACKED, dest, tag, buf->comm, Req);
}

int MMPI_Buf_issend( MMPI_Buf buf, int dest, int tag, MPI_Request * Req )
{
  int R ;

  if ( (buf)->base!=(void *)(((char*)buf)+sizeof(struct MMPI_Buf_type))){
    return MPI_ERR_BUFFER ;
  }

  return MPI_Issend( buf->base, buf->pos, MPI_PACKED, dest, tag, buf->comm, Req);
}

int MMPI_Buf_ibsend( MMPI_Buf buf, int dest, int tag, MPI_Request * Req )
{
  int R ;

  if ( (buf)->base!=(void *)(((char*)buf)+sizeof(struct MMPI_Buf_type))){
    return MPI_ERR_BUFFER ;
  }

  return MPI_Ibsend( buf->base, buf->pos, MPI_PACKED, dest, tag, buf->comm, Req);
}


int MMPI_Buf_recv_init(  MMPI_Buf buf, int src,  int tag, MPI_Request * Req )
{
  int R ;

  if ( (buf)->base!=(void *)(((char*)buf)+sizeof(struct MMPI_Buf_type))){
    return MPI_ERR_BUFFER ;
  }

  return MPI_Recv_init( buf->base, buf->maxsize, MPI_PACKED,
    src, tag, buf->comm, Req );
}

int MMPI_Buf_send_init(  MMPI_Buf buf, int dest, int tag, MPI_Request * Req )
{
  int R ;

  if ( (buf)->base!=(void *)(((char*)buf)+sizeof(struct MMPI_Buf_type))){
    return MPI_ERR_BUFFER ;
  }

  return MPI_Send_init( buf->base, buf->pos, MPI_PACKED,
    dest, tag, buf->comm, Req );
}

int MMPI_Buf_rsend_init( MMPI_Buf buf, int dest, int tag, MPI_Request * Req )
{
  int R ;

  if ( (buf)->base!=(void *)(((char*)buf)+sizeof(struct MMPI_Buf_type))){
    return MPI_ERR_BUFFER ;
  }

  return MPI_Rsend_init( buf->base, buf->pos, MPI_PACKED,
    dest, tag, buf->comm, Req );
}

int MMPI_Buf_ssend_init( MMPI_Buf buf, int dest, int tag, MPI_Request * Req )
{
  int R ;

  if ( (buf)->base!=(void *)(((char*)buf)+sizeof(struct MMPI_Buf_type))){
    return MPI_ERR_BUFFER ;
  }

  return MPI_Ssend_init( buf->base, buf->pos, MPI_PACKED,
    dest, tag, buf->comm, Req );
}

int MMPI_Buf_bsend_init( MMPI_Buf buf, int dest, int tag, MPI_Request * Req )
{
  int R ;

  if ( (buf)->base!=(void *)(((char*)buf)+sizeof(struct MMPI_Buf_type))){
    return MPI_ERR_BUFFER ;
  }

  return MPI_Bsend_init( buf->base, buf->pos, MPI_PACKED,
    dest, tag, buf->comm, Req );
}

/* ------------------------------------------------------------------- */

int MMPI_Buf_verify( MMPI_Buf buffer )
{
  const char my_name[] = "MMPI_Buf_verify" ;

  if ( buffer == (MMPI_Buf) NULL || buffer == MMPI_BUF_NULL ) {
    fprintf(MMPI_Log_file(),"%s : null buffer pointer",my_name);
    fflush(MMPI_Log_file());
    return MPI_ERR_ARG ;
  }

  if ( buffer->base !=
       (void *)(((char*)buffer)+sizeof(struct MMPI_Buf_type))) {
    fprintf(MMPI_Log_file(),"%s : corrupted buffer",my_name);
    fflush(MMPI_Log_file());
    return MPI_ERR_BUFFER ;
  }

  return MPI_SUCCESS ;
}

/* ------------------------------------------------------------------- */

