/* ---------------------------------------------------------------------
Author:     H. Carter Edwards
		http://www.ticam.utexas.edu/~carter
		carter@ticam.utexas.edu

Copyright:  Copyright (C) 1996   H. Carter Edwards

Purpose:    Asynchronous message management for MPI

License:    See 'License.Gnu'

Mail:
	University of Texas at Austin
	TICAM, TAY 2.400
	2313 Speedway
	Austin, TX  78712

Disclaimer:

    MMPI comes with ABSOLUTELY NO WARRANTY; for details read
    the file License.Gnu that came with this library.
    This is free software, and you are welcome to redistribute it
    under certain conditions, i.e. read the file 'License.Gnu'.
--------------------------------------------------------------------- */

#include <stdlib.h>
#include "mmpi.h"

/*
#define DEBUG
*/

/* ------------------------------------------------------------------ */
/* ------------------------------------------------------------------ */

typedef struct {
  MMPI_Request_handler * handler ;
  void                 * extra_state ;
  MPI_Status           * status ;
  MPI_Request          * request ;
} handler_data ;

static int            Waiting = 0 ; /* Waiting request count */
static int            Pending = 0 ; /* Pending request count */
static int            MList   = 0 ; /* Maximum number in list */
static int            NList   = 0 ; /* Current number in list */
static handler_data * DList   = 0 ; /* Handler data list */
static MPI_Request  * RList   = 0 ; /* Request list */
static MPI_Status   * MSList  = 0 ; /* Status list */
static int          * IndList = 0 ; /* Index list */

static double Time_Waiting = 0.0 ;

/* ------------------------------------------------------------------ */

static int handler_default(
  void        * extra_state , 
  MPI_Request * request ,
  MPI_Status  * status )
{
  return MPI_SUCCESS ;
}

/* ------------------------------------------------------------------ */

static int list_find( const MPI_Request Req )
{
  const int NL = NList ;
  const MPI_Request * const RL = RList ;
  register int i ;
  for ( i = NL - 1 ; 0 <= i ; i-- ) if ( RL[i] == Req ) return i ;
  return -1 ;
}

/* ------------------------------------------------------------------ */

static void list_size( const int N )
{
  const int Increment = 64 ;
  const int Size_all =
    sizeof(handler_data) +
    sizeof(MPI_Request) +
    sizeof(MPI_Status) +
    sizeof(int) ;

  handler_data * DL   = DList   ; /* Handler data list */
  MPI_Request  * RL   = RList   ; /* Request list */
  MPI_Status   * MSL  = MSList  ; /* Status list */
  int          * IndL = IndList ; /* Index list */

  const int Len = N + ((N % Increment) ? (Increment - N % Increment) : 0);

  if ( MList < Len ) { /* Enlarge array */

    char * B = malloc( Len * Size_all );

    if ( B == NULL ) {
      MMPI_Log_abort("MMPI Request Manager","Failed malloc",MPI_ERR_UNKNOWN);
    }

    DList   = (handler_data *)   B ;
    RList   = (MPI_Request  *) ( B += Len * sizeof(handler_data) );
    MSList  = (MPI_Status   *) ( B += Len * sizeof(MPI_Request) );
    IndList = (int          *) ( B += Len * sizeof(MPI_Status) );

    if ( MList ) {
      memcpy( DList ,   DL ,   NList * sizeof(handler_data) );
      memcpy( RList ,   RL ,   NList * sizeof(MPI_Request) );
      memcpy( MSList ,  MSL ,  NList * sizeof(MPI_Status) );
      memcpy( IndList , IndL , NList * sizeof(int) );
    }

    free( (char *) DL );
  }
  else if ( Len <= MList / 2 ) {

    if ( Len ) {
      char * B = malloc( Len * Size_all );

      if ( B == NULL ) {
        MMPI_Log_abort("MMPI Request Manager","Failed malloc",MPI_ERR_UNKNOWN);
      }

      DList   = (handler_data *)   B ;
      RList   = (MPI_Request  *) ( B += Len * sizeof(handler_data) );
      MSList  = (MPI_Status   *) ( B += Len * sizeof(MPI_Request) );
      IndList = (int          *) ( B += Len * sizeof(MPI_Status) );

      memcpy( DList ,   DL ,   NList * sizeof(handler_data) );
      memcpy( RList ,   RL ,   NList * sizeof(MPI_Request) );
      memcpy( MSList ,  MSL ,  NList * sizeof(MPI_Status) );
      memcpy( IndList , IndL , NList * sizeof(int) );
    }
    else {
      DList   = 0 ;
      RList   = 0 ;
      MSList  = 0 ;
      IndList = 0 ;
    }

    free( (char *) DL );
  }

  MList = Len ;
}

/* ------------------------------------------------------------------ */

static void list_clean(void)
{
  handler_data * DL = DList ;
  MPI_Request  * RL = RList ;
  int i ;

/* HCE: 3/19/97
  for ( i = 0 ; i < NList &&
    ( DL[i].handler != MMPI_REQUEST_HANDLER_NULL ||
      RL[i] != MPI_REQUEST_NULL); i++ );
*/
  for ( i = 0 ; i < NList && RL[i] != MPI_REQUEST_NULL ; i++ );

  if ( i < NList ) {
    int pop = 1 ;
    NList-- ;

    while ( i < NList ) {
/*
      DL[i] = DL[i+pop] ;
*/
      DL[i].request     = DL[i+pop].request ;
      DL[i].status      = DL[i+pop].status ;
      DL[i].handler     = DL[i+pop].handler ;
      DL[i].extra_state = DL[i+pop].extra_state ;
      RL[i] = RL[i+pop] ;

/* HCE: 3/19/97
      if ( DL[i].handler == MMPI_REQUEST_HANDLER_NULL &&
           RL[i] == MPI_REQUEST_NULL ) {
*/
      if ( RL[i] == MPI_REQUEST_NULL ) {
        pop++ ;
        NList-- ;
      }
      else {
        i++ ;
      }
    }
    list_size( NList );
  }
}

/* ------------------------------------------------------------------ */
/* ------------------------------------------------------------------ */

int MMPI_Post_handler(
  MPI_Request          request ,     /* IN: */
  void               * extra_state , /* IN: */
  MMPI_Request_handler handler )     /* IN: */
{
  if ( request == MPI_REQUEST_NULL ) {
    return MPI_ERR_REQUEST ;
  }
  else {

    int I = list_find( request );

    if ( handler != MMPI_REQUEST_HANDLER_NULL ) {
      if ( I < 0 ) {
        list_size( NList + 1 ); /* Enlarge the list */

        /* Save the information */

        DList[NList].request     = NULL ;
        DList[NList].status      = NULL ;
        DList[NList].handler     = handler ;
        DList[NList].extra_state = extra_state ;
        RList[NList]             = request ;

        NList++ ;
      }
      else {
        DList[I].handler     = handler ;
        DList[I].extra_state = extra_state ;
      }
    }
    else { /* Remove handler from the list */
      if ( 0 <= I ) {
        DList[I].request     = NULL ;
        DList[I].status      = NULL ;
        DList[I].handler     = MMPI_REQUEST_HANDLER_NULL ;
        DList[I].extra_state = NULL ;
        RList[I]             = MPI_REQUEST_NULL ;
        if ( ! Waiting ) list_clean();
      }
    }
  }
  return MPI_SUCCESS ;
}

/* ------------------------------------------------------------------ */
/* The lists may be reallocated during this operation,
   however, the indices will remain fixed
 */

static void serve_pending(void)
{
  while ( Pending ) {

    const int P = --Pending ;
    const int I = IndList[ P ];
    MPI_Request Req  = RList[I] ;
    MPI_Status  Stat = MSList[P] ;

    /* If testing or waiting for this particular request,
       Output the request state and status.
     */

    if ( DList[I].request )
      { *(DList[I].request) = Req ; DList[I].request = NULL ; }

    if ( DList[I].status  )
      { *(DList[I].status) = Stat ; DList[I].status = NULL ; }

#ifdef DEBUG
  if ( DList[I].handler ) {
    fprintf(MMPI_Log_file(),"\n  handler%d{%d,%d}\n",
      I,
      Stat.MPI_SOURCE,
      Stat.MPI_TAG);
  }
  else {
    fprintf(MMPI_Log_file(),"\n  no-handler%d{%d,%d}\n",
      I,
      Stat.MPI_SOURCE,
      Stat.MPI_TAG);
  }
  fflush(MMPI_Log_file());
#endif

    if ( DList[I].handler ) {

      const int R = (*DList[I].handler)( DList[I].extra_state, &Req, &Stat);

      if ( MPI_SUCCESS != R ) {
        MMPI_Log_abort("MMPI Request Manager","Failed Handler",R);
      }

      /* These may have been updated and the list may have changed */

      RList[I]  = Req ;
      MSList[P] = Stat ;
    }
  }
}

/* ------------------------------------------------------------------ */

double MMPI_Wait_idle_time(void) { return Time_Waiting ; }

int MMPI_Wait( MPI_Request * request , MPI_Status * status )
{
  int J ;
  int R ;
  int P = 0 ;

  serve_pending();

  J = list_find( *request );

  if ( J < 0 ) {
    P = 1 ;
    J = NList ;

    list_size( NList + 1 ); /* Enlarge the list */

    DList[J].handler     = MMPI_REQUEST_HANDLER_NULL ;
    DList[J].extra_state = status ;
    RList[J]             = *request ;

    NList++ ;
  }

  DList[J].request = request ; /* Wait for this request */
  DList[J].status  = status ;

  Waiting++ ; /* Lock the list indices */

  do {

     double T = MPI_Wtime();

     R = MPI_Waitsome(NList,RList,&Pending,IndList,MSList);

     Time_Waiting += MPI_Wtime() - T ;

     if ( MPI_SUCCESS == R ) serve_pending();

  } while ( MPI_SUCCESS == R && DList[J].request ); /* Wait for request */

  Waiting-- ; /* Unlock the list indices */

  if ( P ) {
    DList[J].request     = NULL ;
    DList[J].status      = NULL ;
    DList[J].handler     = MMPI_REQUEST_HANDLER_NULL ;
    DList[J].extra_state = NULL ;
    RList[J]             = MPI_REQUEST_NULL ;
  }

  if ( ! Waiting ) list_clean();

  return ( MPI_SUCCESS == R ) ? MMPI_Serve_handler() : R ;
}

/* ------------------------------------------------------------------ */

int MMPI_Test( MPI_Request * request , int * flag , MPI_Status * status )
{
  int J ;
  int R ;
  int P = 0 ;

  serve_pending();

  J = list_find( *request );

  if ( J < 0 ) {
    P = 1 ;
    J = NList ;

    list_size( NList + 1 ); /* Enlarge the list */

    DList[J].handler     = MMPI_REQUEST_HANDLER_NULL ;
    DList[J].extra_state = NULL ;
    RList[J]             = *request ;

    NList++ ;
  }

  DList[J].request = request ;
  DList[J].status  = status ;

  Waiting++ ; /* 'Lock' the list indices */

  R = MPI_Testsome(NList,RList,&Pending,IndList,MSList);

  if ( MPI_SUCCESS == R ) serve_pending();

  Waiting-- ; /* 'Unlock' the list indices */

  *flag = ( DList[J].status == NULL ) ? 1 : 0 ;

  if ( P ) {
    DList[J].request     = NULL ;
    DList[J].status      = NULL ;
    DList[J].handler     = MMPI_REQUEST_HANDLER_NULL ;
    DList[J].extra_state = NULL ;
    RList[J]             = MPI_REQUEST_NULL ;
  }

  if ( ! Waiting ) list_clean();

  return ( MPI_SUCCESS == R ) ? MMPI_Serve_handler() : R ;
}

/* ------------------------------------------------------------------ */

int MMPI_Serve_handler(void)
{
  if ( NList ) {
    int R ;

    Waiting++ ; /* Lock the list indices */

    do {

      serve_pending();

      if ( MPI_SUCCESS != ( R =
           MPI_Testsome(NList,RList,&Pending,IndList,MSList) ) ) return R ;

    } while( Pending );

    Waiting-- ; /* Unlock the list indices */

    if ( ! Waiting ) list_clean();
  }

  return MPI_SUCCESS ;
}

/* ------------------------------------------------------------------ */

