#include <mpi.h>
#include <PLA.h>
#include <MISC_DEFS.h>
#include "PRISM.h"

void PRISM_split_into_subproblems(PLA_Obj X, int n_1, PLA_Obj Q, 
     PLA_Obj X_0, PLA_Obj X_1) {
/*
     Split into subproblems.

     Method: 
       T = Q' * X * Q
       X_1 = upper left block of T 
       X_0 = lower right block of T 
     
     where:
       X_1 = subproblem corrpesonding to the range space of X 
       X_0 = subproblem corresponding to the null space of X
*/

    int           length, align_row, align_col;
    PLA_Obj       T=NULL, T_0=NULL, T_1=NULL, Q_0=NULL, Q_1=NULL;
    PLA_Obj       DB_0=NULL, DB_1=NULL;
    MPI_Datatype  datatype;
    PLA_Template  template;

    /* Create temporary matrix */
    PLA_Matrix_create_conf_to( X, &T );

    /* First part: T = Q' * T */ 
    PLA_Obj_set_to_zero(T);
    PLA_Gemm( PLA_TRANS, PLA_NO_TRANS, PRISM_one, Q, X, PRISM_zero, T );

    /* Save ops by only calculating blocks needed */
    PLA_Obj_horz_split_2( T, n_1, &T_1, &T_0 );
    PLA_Obj_vert_split_2( Q, n_1, &Q_1, &Q_0 ); 

    /* Second part: X_0 = T_0 * Q_0, X_1 = T_1 * Q_1 */
    PLA_Obj_set_to_zero(X_1);
    PLA_Obj_set_to_zero(X_0);
    PLA_Gemm( PLA_NO_TRANS, PLA_NO_TRANS, PRISM_one, T_1, Q_1, PRISM_zero, X_1 );
    PLA_Gemm( PLA_NO_TRANS, PLA_NO_TRANS, PRISM_one, T_0, Q_0, PRISM_zero, X_0 );

    /* Symmetrize X_0 and X_1 */
    PRISM_sym(X_1);
    PRISM_sym(X_0);

#ifdef DB_GLOBAL_
    if (me==ROOT) printf(">> PRISM_split_into_subproblems\n");
    PRISM_print(X_1); 
    PRISM_print(X_0); 

    /* Compute off-diagonal blocks */
    PLA_Obj_global_length	  ( X, &length );
    PLA_Obj_global_align_row  ( X, &align_row );
    PLA_Obj_global_align_col  ( X, &align_col );
    PLA_Obj_datatype		  ( X, &datatype );
    PLA_Obj_template		  ( X, &template );

    PLA_Matrix_create(datatype, n_1, length-n_1, template, align_row, 
       align_col+n_1, &DB_0);
    PLA_Matrix_create(datatype, length-n_1, n_1, template, align_row+n_1, 
       align_col, &DB_1);

    PLA_Obj_set_to_zero(DB_0);
    PLA_Obj_set_to_zero(DB_1);
    PLA_Gemm( PLA_NO_TRANS, PLA_NO_TRANS, PRISM_one, T_1, Q_0, PRISM_zero, DB_0 );
    PLA_Gemm( PLA_NO_TRANS, PLA_NO_TRANS, PRISM_one, T_0, Q_1, PRISM_zero, DB_1 );

    PRISM_print(DB_0); 
    PRISM_print(DB_1); 

    PLA_Obj_free(&DB_0);
    PLA_Obj_free(&DB_1);
#endif

    PLA_Obj_free(&T);
    PLA_Obj_free(&T_1); 
    PLA_Obj_free(&T_0);
    PLA_Obj_free(&Q_1); 
    PLA_Obj_free(&Q_0);
}

