/* Author:      Sunlung Suen (ssuen@cs.wisc.edu)
 * Class:       CS757   University of Wisconsin-Madison
 * Instructor:  Dr. James Goodman
 * Assignment:  Homework 2
 * Due Date:    2/28/1997
 */

/********************
 * Parallel Version *
 ********************/

/* Program Description:  This program finds the Nth prime number, where N
 *                       is interactively entered by the user.
 */

/* Target Platform:  This program is intended to run on a network of 
 *                   workstations supporting MPI.
 */

/* Algorithm:  Multiple copies of this program should first be started on 
 *             multiple workstations.  The processes will then logically 
 *             arrange themselves into a ring, with each process having a left
 *             and a right neighbor.  Each process then repeatedly takes a 
 *             range of four thousand integers, finds all the primes in the 
 *             range, receives the current primes-found count from its left 
 *             neighbor, checks if the prime being calculated has been found,
 *             updates the count with the number of primes it just found, and
 *             sends the updated count to its right neighbor.
 */

/* Observations:  Except for 2, all primes must be odd.  Nearly half the odd
 *                numbers are multiples of 3 and/or 5.
 */


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "mpi.h"


unsigned long findPrime();     /*Parallel prime number search function       */
int           isPrime();       /*Boolean function to test if integer is prime*/
unsigned long getWhichPrime(); /*Asks user which prime he/she wants          */


int main(argc, argv)
         int    argc;
         char **argv;
{
  int            numProcs;          /*Number of processes started       */
  int            myID;              /*Identifier of this process        */
  unsigned long  whichPrime;        /*Which prime to search for?        */
  double         startTime;         /*Time at start of prime search     */
  unsigned long  primeNum;          /*The prime number found            */
  void          *sendBuf;           /*Used for nonblocking message sends*/
  int            sendBufSize = 128; /*Size of nonblocking message buffer*/

  MPI_Init(&argc, &argv);                     /*Initialize MPI environment   */
  MPI_Comm_size(MPI_COMM_WORLD, &numProcs);   /*How many processes started?  */
  MPI_Comm_rank(MPI_COMM_WORLD, &myID);       /*What is this process' ID?    */
                                              /*Enable nonblocking sends     */
  MPI_Buffer_attach(malloc(sendBufSize), sendBufSize);

  while (1) {                                 /*Until user wants to quit     */
    if (myID == 0) {                          /*All I/O done by local process*/
      whichPrime = getWhichPrime();
      startTime  = MPI_Wtime();
    }
                               /*Let all processes know which prime is needed*/
    MPI_Bcast(&whichPrime, 1, MPI_UNSIGNED_LONG, 0, MPI_COMM_WORLD);

    if (whichPrime == 0) break;               /*User wants to quit?          */

    primeNum = findPrime(whichPrime, myID, numProcs);

    if (myID == 0) {
      printf("The %luth prime is %lu\n", whichPrime, primeNum);
      printf("Wall clock time is %f seconds\n", MPI_Wtime() - startTime);
    }
  }

  MPI_Buffer_detach(&sendBuf, &sendBufSize);  /*Free nonblocking message buf */
  free(sendBuf);

  MPI_Finalize();                             /*Cleanup MPI environment      */
}


#define PS 4000  /*Partition size -- processes work on ranges of 4000 ints*/

/* Observations:  The value of the Nth prime is greater than N.
 *                At most 1/4 the integers in a range of 4000 are prime.
 */

              /*Parallel prime number search function, returns the prime found
               */
unsigned long findPrime(whichPrime, myID, numProcs)
                        unsigned long whichPrime; /*Which prime is needed?   */
                        int           myID;       /*ID for this process      */
                        int           numProcs;   /*Total Number of processes*/
{  
  unsigned long partition;    /*The range of integers to search              */
  unsigned long primes[PS/4]; /*Primes found in the searched range           */
  unsigned long found;        /*Number of primes found in searched range     */
  unsigned long i;            /*Prime integer candidate                      */
  unsigned long u;            /*Upper limit of range being searched          */
  unsigned long pf = 0;       /*Total number of primes found by all processes*/
                              /*  (and, when it is found, the prime found)   */
  MPI_Status    status;       /*Status of received message                   */
 
  partition = (unsigned long) myID;                    /*First range to check*/
 
  while (whichPrime > pf) {                            /*Still searching     */
    if (partition)
      found = 0;
    else {                                     /*1st primes are special cases*/
      primes[0] = 2; primes[1] = 3;  primes[2] = 5;
      primes[3] = 7; primes[4] = 11; primes[5] = 13;
      found = 6;
    }

    switch ((partition * PS) % 3) {               /*Set first integer to test*/
      case 0:  i = partition * PS + 15; break;
      case 1:  i = partition * PS + 5;  break;
      case 2:  i = partition * PS + 25; break;
    }                                             /*i is divisible by 3 and 5*/

    switch (((partition + 1) * PS) % 3) {   /*Set upper limit of search range*/
      case 0:  u = (partition + 1) * PS + 15; break;
      case 1:  u = (partition + 1) * PS + 5;  break;
      case 2:  u = (partition + 1) * PS + 25; break;
    }                                             /*u is divisible by 3 and 5*/

      /*Test odd integers in search range, skipping those divisible by 3 or 5*/
    for (i = i + 2; i < u; i += 4) {
      if (isPrime(i)) { primes[found] = i; found++; }
      i += 2; if (isPrime(i)) { primes[found] = i; found++; }
      i += 4; if (isPrime(i)) { primes[found] = i; found++; }
      i += 6; if (isPrime(i)) { primes[found] = i; found++; }
      i += 2; if (isPrime(i)) { primes[found] = i; found++; }
      i += 6; if (isPrime(i)) { primes[found] = i; found++; }
      i += 4; if (isPrime(i)) { primes[found] = i; found++; }
      i += 2; if (isPrime(i)) { primes[found] = i; found++; }
    }
                      /*Receive current primes-found count from left neighbor*/
    if (partition)
      MPI_Recv(&pf, 1, MPI_UNSIGNED_LONG, (myID ? myID - 1 : numProcs - 1),
               MPI_ANY_TAG, MPI_COMM_WORLD, &status);

    if ((pf + found) < whichPrime) {           /*Prime not found             */
      pf += found;        /*Update primes-found count; send to right neighbor*/
      MPI_Bsend(&pf, 1, MPI_UNSIGNED_LONG, (myID + 1) % numProcs, 1, 
                MPI_COMM_WORLD);
      partition += numProcs;                   /*Next integer range to check */
    }
    else if (pf > whichPrime)  /*Prime already found; let right neighbor know*/
      MPI_Send(&pf, 1, MPI_UNSIGNED_LONG, (myID + 1) % numProcs, 1, 
               MPI_COMM_WORLD);
    else {                                     /*Just found prime            */
      pf = primes[whichPrime - pf - 1];        /*Let right neighbor know     */
      MPI_Send(&pf, 1, MPI_UNSIGNED_LONG, (myID + 1) % numProcs, 1, 
               MPI_COMM_WORLD);           /*Wait for message to traverse ring*/
      MPI_Recv(&pf, 1, MPI_UNSIGNED_LONG, (myID ? myID - 1 : numProcs - 1),
               MPI_ANY_TAG, MPI_COMM_WORLD, &status);          
    }
  }
  
  return(pf);                                  /*Return prime found          */
}



    /* Boolean function to test if integer is prime.  Integer being tested
     *   MUST be 
     *     1.  Odd
     *     2.  Not a multiple of 3 or 5
     *     3.  Greater than 15
     */
int isPrime(i)
            unsigned long i;
{
  unsigned long d; /*Divisors to try       */
  unsigned long s; /*Square root of integer*/

  if (i%7==0 || i%11==0 || i%13==0)
    return(0);

  s = (unsigned long) sqrt((double) i);

            /*Try dividing by odd numbers, skipping those divisible by 3 or 5*/
  for (d = 17; d <= s; d += 4) {
    if (i % d == 0) return(0);
    d += 2; if (i % d == 0) return(0);
    d += 4; if (i % d == 0) return(0);
    d += 6; if (i % d == 0) return(0);
    d += 2; if (i % d == 0) return(0);
    d += 6; if (i % d == 0) return(0);
    d += 4; if (i % d == 0) return(0);
    d += 2; if (i % d == 0) return(0);
  }
  return(1);
}


              /*Asks user which prime he/she wants; returns what user inputs
               */
unsigned long getWhichPrime()
{
  long whichPrime;    /*Which prime is needed?*/


  while (1) {                               /*Until valid input*/
    printf("\nPlease enter which prime (0 quits):\n");

    if (scanf(" %ld", &whichPrime) != 1)
      while (getchar() != '\n') ;           /*Discard bad input*/
    else if (whichPrime >= 0)
      return((unsigned long) whichPrime);

    printf("Sorry, must be a non-negative number.\n");
  }
}
